import numpy as np
import json
import ast


class Ensemble(object):
    """An ensemble of particles.

    All ensembles have particle positions and velocities. In addition, an
    ensemble may define ensemble properties and per particle properties."""

    def __init__(self, num_ptcls=1):
        self.x = np.zeros([num_ptcls, 3], dtype=np.float64)
        self.v = np.zeros([num_ptcls, 3], dtype=np.float64)
        self.ensemble_properties = {}
        self.particle_properties = {}

    def get_num_ptcls(self):
        return self.x.shape[0]

    num_ptcls = property(get_num_ptcls,
                         'The number of particles in the ensemble')

    def set_particle_property(self, key, prop):
        """Set a particle property.

        key -- The name under which the property is stored.
        prop -- An array of the per particle properties. The array should have
        layout (num_ptcls, ...), i.e. the leading dimension should match the
        number of particles currently in the ensemble. We store a copy of
        prop."""

        if prop.shape[0] != self.num_ptcls:
            raise RuntimeError(
                'Size of property array does not match number of particles.')
        self.particle_properties[key] = np.copy(prop)

    def copy(self):
        the_copy = Ensemble(self.num_ptcls)
        the_copy.x = np.copy(self.x)
        the_copy.v = np.copy(self.v)
        the_copy.ensemble_properties = dict(self.ensemble_properties)
        for key in self.particle_properties:
            the_copy.particle_properties[key] = np.copy(self.particle_properties[key])
        return the_copy

    def resize(self, new_size):
        shape = list(self.x.shape)
        shape[0] = new_size
        self.x.resize(shape)
        self.v.resize(shape)
        for particle_prop in self.particle_properties:
            shape = [self.particle_properties[particle_prop].shape]
            shape[0] = new_size
            self.particle_properties[particle_prop].resize(shape)

    def delete(self, indices):
        """Delete a subset of particles.

        The particles will be deleted inplace, i.e. the ensemble is mutated.

        indices -- The indices of particles to delete.
        """
        self.x = np.delete(self.x, indices, 0)
        self.v = np.delete(self.v, indices, 0)
        for prop in self.particle_properties:
            prop = np.delete(prop, indices, 0)


def ensemble_to_json(ensemble):
    """Serialize an Ensemble to a JSON string"""
    serialized = {}
    serialized['x'] = json.dumps(ensemble.x.tolist())
    serialized['v'] = json.dumps(ensemble.v.tolist())
    serialized['ensemble_properties'] = ensemble.ensemble_properties
    particle_properties = {}
    for prop,val in ensemble.particle_properties.items():
        particle_properties[prop] = json.dumps(val.tolist())
    serialized['particle_properties'] = particle_properties
    return json.dumps(serialized)


def json_to_ensemble(s):
    """Deserialize an Ensemble from a JSON string"""
    d = json.loads(s)
    x = np.array(ast.literal_eval(d['x']))
    v = np.array(ast.literal_eval(d['v']))
    ensemble_properties = d['ensemble_properties']
    particle_properties = {}
    for key in d['particle_properties']:
        val = d['particle_properties'][key]
        particle_properties[key] = np.array(ast.literal_eval(val))
    num_ptcls = x.shape[0]
    ensemble = Ensemble(num_ptcls)
    ensemble.x = x
    ensemble.v = v
    ensemble.ensemble_properties = ensemble_properties
    ensemble.particle_properties = particle_properties
    return ensemble


class Source(object):
    """A particle source."""

    def __init__(self):
        pass

    def num_ptcls_produced(self, dt):
        """The number of particles that will be produced.

        This gives the number of particles that will be generated by the next
        call to produce_ptcls. It is legal for a source to return different
        numbers of particles on consecutive calls to num_ptcls_produced. This
        is often the case for stochastic sources that produce a certain number
        of particles on average.

        dt -- The duration of the time interval for which to produce particles.
        """
        return 0

    def produce_ptcls(self, dt, start, end, ensemble):
        """Generate particles.

        The number of particles generated must match the result return by a
        call to num_ptcls_produced(). We can expect end - start ==
        num_ptcls_produced().

        dt -- The duration of the time interval for which to produce particles.
        To make start and end and dt consistent with one another this dt should
        be the same as the dt passed to num_ptcls_produced to obtain start and
        end.
        start -- First position in ensemble where particles will be inserted.
        end -- One past the last position in the ensemble where particles will
        be inserted.
        ensemble -- The ensemble into which to insert the particles."""
        pass


def produce_ptcls(dt, ensemble, sources=[]):
    """Insert particles produced by sources into the ensemble.

    dt -- Length of time interval for which to produce particles.
    ensemble -- The ensemble into which to insert the particles.
    sources -- The particle source. Should derive from Source.
    """

    num_new_ptcls = []
    tot_new_ptcls = 0
    for s in sources:
        num_new_ptcls.append(s.num_ptcls_produced(dt))
        tot_new_ptcls += num_new_ptcls[-1]

    start = ensemble.num_ptcls
    ensemble.resize(ensemble.num_ptcls + tot_new_ptcls)
    for i, s in enumerate(sources):
        s.produce_ptcls(dt, start, start + num_new_ptcls[i], ensemble)
        start += num_new_ptcls[i]


class Sink(object):
    """A particle sink.

    Conceptually, sinks are represented by surfaces that remove particles from
    an ensemble if they hit the surface."""

    def find_absorption_time(self, x, v, dt):
        """The time at which particles will be arbsorbed by the sink.

        This method returns the time interval after which particles starting at
        x and traveling with velocity vector v along a straight line will hit
        the sink surface. dt is the duration of the interval in which an
        absorption time is sought. If the particle will not hit the sink in
        interval dt the function should return an absorption time greater than
        dt.

        x -- Initial particle positions.
        v -- Particle velocities.
        dt -- Length of time interval.
        """
        return np.full(x.shape[0], 2.0 * dt)

    def record_absorption(self, ensemble, dt, absorption_times, absorption_indices):
        """This function gets called when this sink absorbs a particle.

        ensemble -- The ensemble with particles which are potentially absorbed by this sink.
        dt -- The time interval during which we are processing decay events.
        absorption_times -- The times at which particles may be absorbed.
        absorption_indices -- The indices of particles which are being absorbed by this sink.
        """
        pass


class SinkPlane(Sink):

    def __init__(self, point, normal):
        """Generate a sink that absorbs particles hitting a plane.

        point -- A point in the plane.
        normal -- A normal to the plane.
        """
        self.point = point
        self.normal = normal

    def find_absorption_time(self, x, v, dt):
        taus = np.empty(x.shape[0])

        for i in range(x.shape[0]):
            normal_velocity = self.normal.dot(v[i])
            if (normal_velocity == 0.0):
                taus[i] = 2.0 * dt
            else:
                taus[i] = self.normal.dot(self.point - x[i]) / normal_velocity

        return taus


def process_sink(dt, ensemble, sink):
    if sink == None:
        return
    absorption_times = sink.find_absorption_time(ensemble.x, ensemble.v, dt)
    absorption_indices = np.arange(ensemble.num_ptcls)[
        abs(absorption_times - 0.5 * dt) <= 0.5 * dt]
    sink.record_absorption(ensemble, dt, absorption_times, absorption_indices)
    ensemble.delete(absorption_indices)


def drift_kick(dt, ensemble, forces=[], sink=None):
    """Drift-Kick-Drift push of particles.

    dt --       Time step size.
    ensemble -- The ensemble to advance.
    forces --   Forces acting on the ensemble. Each entry in this list must have
                a method force(dt, ensemble, f) that adds the force integrated
                over dt to f.
    sink --     The particle sink.
    
    """
    if len(forces) == 0:
        process_sink(dt, ensemble, sink)
        ensemble.x += dt * ensemble.v
    else:
        process_sink(0.5 * dt, ensemble, sink)

        ensemble.x += 0.5 * dt * ensemble.v

        f = np.zeros_like(ensemble.v)
        for force in forces:
            force.force(dt, ensemble, f)

        m = 0.0
        if 'mass' in ensemble.ensemble_properties:
            m = ensemble.ensemble_properties['mass']
        elif 'mass' in ensemble.particle_properties:
            m = ensemble.particle_properties['mass']
        else:
            raise RuntimeError('To accelerate particles we need a mass ensemble or particle property')
        ensemble.v +=  f / m

        process_sink(0.5 * dt, ensemble, sink)

        ensemble.x += 0.5 * dt * ensemble.v




import numpy as np
import coldatoms_lib


def _coulomb_force_ref(positions, q, dt, delta, k, f):
    kp = dt * k * q * q
    num_ptcls = positions.shape[0]
    for i in range(num_ptcls):
        for j in range(num_ptcls):
            r = positions[i] - positions[j]
            absr = np.sqrt(r.dot(r) + delta)
            f[i] += kp * r / (absr * absr * absr)


def _coulomb_force_ref_per_particle_charges(positions, q, dt,
                                            delta, k, f):
    num_ptcls = positions.shape[0]
    k *= dt
    for i in range(num_ptcls):
        for j in range(num_ptcls):
            r = positions[i] - positions[j]
            absr = np.sqrt(r.dot(r) + delta)
            f[i] += k * q[i] * q[j] * r / (absr * absr * absr)


class CoulombForce(object):

    _epsilon0 = 8.854187817620e-12
    _k = 1.0 / (4.0 * np.pi * _epsilon0)

    def __init__(self):
        self.delta = 0.0
        self.coulomb_force = coldatoms_lib.coulomb_force
        self.coulomb_force_per_particle_charges = coldatoms_lib.coulomb_force_per_particle_charge

    def use_reference_implementations(self):
        self.coulomb_force = _coulomb_force_ref
        self.coulomb_force_per_particle_charges = _coulomb_force_ref_per_particle_charges

    def force(self, dt, ensemble, f):
        positions = ensemble.x

        # We guard against the case where particles are in the same
        # location by adding a small cut off parameter to the
        # softcore parameter. This cut off parameter is designed so that
        # the distance raised to the third power is still different
        # from zero so we can savely divide.
        ulp = np.finfo(ensemble.x.dtype).tiny
        my_delta_squared = (self.delta * self.delta +
                            1.0e1 * ulp **(2.0/3.0))

        if 'charge' in ensemble.ensemble_properties:
            q = ensemble.ensemble_properties['charge']
            self.coulomb_force(positions, q, dt,
                               my_delta_squared, self._k, f)
        elif 'charge' in ensemble.particle_properties:
            q = ensemble.particle_properties['charge']
            self.coulomb_force_per_particle_charges(
                positions, q, dt, my_delta_squared, self._k, f)
        else:
            raise RuntimeError('Must provide a charge to compute coulomb force')




from .coldatoms_lib import (
    compute_nbars,
    add_radiation_pressure,
    coulomb_force,
    coulomb_force_per_particle_charge,
    harmonic_trap_forces,
    harmonic_trap_forces_per_particle_charge,
    bend_kick_update,
    Rng,
    rng
    )




import numpy as np
import coldatoms_lib


class RadiationPressure(object):
    """The force experienced by level atoms undergoing resonance fluorescence.

    This class computes the radiation pressure force, both determinstic and
    fluctuating recoil components, on an atom driven by a monochromatic laser
    field. The class handles spatial variations in laser intensity and of the
    atomic transition frequency as in a Zeeman slower. The wavevector of the
    laser field is assumed constant. RadiationPressure does not deal with
    attenuation of the laser field and is therefore limited to optically thin
    samples.
    An important application of RadiationPressure is to the simulation of
    Doppler cooling of atoms by combining two red detuned lasers with opposite
    propagation directions. In the context of laser cooling we are limited to
    situations of low total saturation and we cannot handle sub-Doppler
    cooling.
    """

    def __init__(self, gamma, hbar_k, intensity, detuning):
        """Specification of a RadiationPressure object.

        gamma -- The atomic decay rate (2\pi / excited state lifetime).
        hbar_k -- Single photon recoil momentum.
        intensity -- The intensity (in units of the saturation intensity) as a
                     function of atomic position. This object must have a
                     method intensities that, when applied to the atomic
                     positions returns the intensities at the position of the
                     atoms.
        detuning -- The detuning of the atomic transition from the laser. Red
                    detuning is negative and blue detuning is positive. The
                    detuning is a function of atomic velocities and may depend
                    on their position. This function must be applicable to x
                    and v and it should return an array with the detuning
                    values.
        """
        self.gamma = gamma
        self.hbar_k = np.copy(hbar_k)
        self.intensity = intensity
        self.detuning = detuning

    def force(self, dt, ensemble, f):
        s_of_r = self.intensity.intensities(ensemble.x)
        deltas = self.detuning.detunings(ensemble.x, ensemble.v)
        nbars = np.zeros_like(deltas)
        coldatoms_lib.compute_nbars(dt, self.gamma, s_of_r, deltas, nbars)
        coldatoms_lib.add_radiation_pressure(coldatoms_lib.rng.context(), self.hbar_k, nbars, f)




import numpy as np
import coldatoms_lib


def bend_kick_update_reference_impl(dt, omegaB, x, v):
    theta = dt * omegaB
    cosTheta = np.cos(theta)
    sinTheta = np.sin(theta)

    x[:, 0] = x[:, 0] + (
        sinTheta * v[:, 0] + (cosTheta - 1.0) * v[:, 1]) / omegaB
    x[:, 1] = x[:, 1] + (
        -(cosTheta - 1.0) * v[:, 0] + sinTheta * v[:, 1]) / omegaB

    vx = cosTheta * v[:, 0] - sinTheta * v[:, 1]
    v[:, 1] = sinTheta * v[:, 0] + cosTheta * v[:, 1]
    v[:, 0] = vx


def bend_kick(dt, Bz, ensemble, forces, num_steps=1, reference_impl=False):
    """Integrator for particle motion in a constant magnetic field"""

    if num_steps < 1:
        return

    if 'charge' in ensemble.ensemble_properties:
        q = ensemble.ensemble_properties['charge']
    elif 'charge' in ensemble.particle_properties:
        q = ensemble.particle_properties['charge']
    else:
        raise RuntimeError('Must provide a charge for bend_kick update.')

    if 'mass' in ensemble.ensemble_properties:
        m = ensemble.ensemble_properties['mass']
    elif 'mass' in ensemble.particle_properties:
        m = ensemble.particle_properties['mass']
    else:
        raise RuntimeError('Must provide a mass for bend_kick update.')

    omegaB = Bz * q / m

    if reference_impl:
         updater = bend_kick_update_reference_impl
    else:
         updater = coldatoms_lib.bend_kick_update

    if len(forces) == 0:
        updater(num_steps * dt, omegaB, ensemble.x, ensemble.v)
    else:
        updater(0.5 * dt, omegaB, ensemble.x, ensemble.v)

        f = np.zeros_like(ensemble.v)

        for i in range(num_steps - 1):
            for force in forces:
                force.force(dt, ensemble, f)
            ensemble.v += f / m
            updater(dt, omegaB, ensemble.x, ensemble.v)
            f.fill(0.0)

        for force in forces:
            force.force(dt, ensemble, f)
        ensemble.v += f / m
        updater(0.5 * dt, omegaB, ensemble.x, ensemble.v)



from .particles import (
    drift_kick,
    process_sink,
    produce_ptcls,
    Ensemble,
    json_to_ensemble,
    ensemble_to_json,
    Sink,
    SinkPlane,
    Source,
    )
from .coulomb import (
    CoulombForce
    )
from .bend_kick import (
    bend_kick
)
from .radiation_pressure import (
    RadiationPressure
)
from .penning_trap import (
    HarmonicTrapPotential
)
from .version import (
    __version__
)
from coldatoms_lib import (
    Rng
)



from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import numpy
import sys
import os


with open('README.md') as f:
    readme = f.read()


with open('LICENSE') as f:
    license = f.read()


exec(open('src/coldatoms/version.py').read())


extra_compile_args = [
    '-DDSFMT_MEXP=19937'
    ]
if 'win' in sys.platform:
    if os.environ['PYTHON_ARCH'] == '64':
        extra_compile_args += [
            '-DHAVE_SSE2'
            ]
else:
    extra_compile_args += [
        '-DHAVE_SSE2',
        '-std=c99',
        '-ffast-math',
        '-ftree-vectorize',
        '-msse4']

extra_link_args = []

coldatoms_lib = cythonize([Extension(
    'coldatoms_lib.coldatoms_lib',
    sources=['src/coldatoms_lib/bend_kick_updater.c',
             'src/coldatoms_lib/forces.c',
             'src/coldatoms_lib/ca_rand.c',
             'src/coldatoms_lib/radiation_pressure.c',
             'src/coldatoms_lib/dSFMT/dSFMT.c',
             'src/coldatoms_lib/coldatoms_lib.pyx'],
    include_dirs=['src/coldatoms_lib/', numpy.get_include()],
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args
    )])

packages = find_packages(where='src',
                         exclude=('tests', 'docs', 'examples'))

setup(
    name='coldatoms',
    version=__version__,
    description='Collection of python tools for cold atoms simulations',
    long_description=readme,
    author='Dominic Meiser',
    author_email='dmeiser79@gmail.com',
    url='https://github.com/d-meiser/cold-atoms',
    license=license,
    packages=packages,
    package_dir={'': 'src'},
    setup_requires=[
        'cython',
        'numpy',
    ],
    install_requires=[
        'cython',
        'numpy',
    ],
    ext_modules=coldatoms_lib
)



import numpy as np
import coldatoms_lib


def _harmonic_trap_forces_ref(positions, q, kx, ky, kz, phi, dt, f):
    cphi = np.cos(phi)
    sphi = np.sin(phi)

    x = positions[:, 0]
    y = positions[:, 1]
    z = positions[:, 2]

    f[:, 0] += dt * q * (
        (-kx * cphi * cphi - ky * sphi * sphi) * x + cphi * sphi * (ky - kx) * y)
    f[:, 1] += dt * q * (
        cphi * sphi * (ky - kx) * x + (-kx * sphi * sphi - ky * cphi * cphi) * y)
    f[:, 2] += -dt * q * kz * z


class HarmonicTrapPotential(object):
    def __init__(self, kx, ky, kz):
        self.kx = kx
        self.ky = ky
        self.kz = kz
        self.phi = 0.0
        self._harmonic_trap_forces = coldatoms_lib.harmonic_trap_forces
        self._harmonic_trap_forces_per_particle_charges = coldatoms_lib.harmonic_trap_forces_per_particle_charge

    def use_reference_implementations(self):
        self._harmonic_trap_forces = _harmonic_trap_forces_ref
        self._harmonic_trap_forces_per_particle_charges = _harmonic_trap_forces_ref

    def force(self, dt, ensemble, f):
        positions = ensemble.x
        if 'charge' in ensemble.ensemble_properties:
            q = ensemble.ensemble_properties['charge']
            self._harmonic_trap_forces(positions, q, self.kx, self.ky, self.kz,
                self.phi, dt, f)
        elif 'charge' in ensemble.particle_properties:
            q = ensemble.particle_properties['charge']
            self._harmonic_trap_forces_per_particle_charges(
                positions, q, self.kx, self.ky, self.kz, self.phi, dt, f)
        else:
            raise RuntimeError('Must provide a charge to compute coulomb force')
